@@ -1,6 +1,3 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
 from itertools import chain
 
 from django.apps import apps
@@ -11,16 +8,19 @@ from django.contrib.admin.utils import (
 from django.core import checks
 from django.core.exceptions import FieldDoesNotExist
 from django.db import models
+from django.db.models.constants import LOOKUP_SEP
 from django.forms.models import (
     BaseModelForm, BaseModelFormSet, _get_foreign_key,
 )
 from django.template.engine import Engine
 
 
-def check_admin_app(**kwargs):
-    from django.contrib.admin.sites import system_check_errors
-
-    return system_check_errors
+def check_admin_app(app_configs, **kwargs):
+    from django.contrib.admin.sites import all_sites
+    errors = []
+    for site in all_sites:
+        errors.extend(site.check(app_configs))
+    return errors
 
 
 def check_dependencies(**kwargs):
@@ -62,10 +62,11 @@ def check_dependencies(**kwargs):
     return errors
 
 
-class BaseModelAdminChecks(object):
+class BaseModelAdminChecks:
 
     def check(self, admin_obj, **kwargs):
         errors = []
+        errors.extend(self._check_autocomplete_fields(admin_obj))
         errors.extend(self._check_raw_id_fields(admin_obj))
         errors.extend(self._check_fields(admin_obj))
         errors.extend(self._check_fieldsets(admin_obj))
@@ -80,6 +81,61 @@ class BaseModelAdminChecks(object):
         errors.extend(self._check_readonly_fields(admin_obj))
         return errors
 
+    def _check_autocomplete_fields(self, obj):
+        """
+        Check that `autocomplete_fields` is a list or tuple of model fields.
+        """
+        if not isinstance(obj.autocomplete_fields, (list, tuple)):
+            return must_be('a list or tuple', option='autocomplete_fields', obj=obj, id='admin.E036')
+        else:
+            return list(chain.from_iterable([
+                self._check_autocomplete_fields_item(obj, obj.model, field_name, 'autocomplete_fields[%d]' % index)
+                for index, field_name in enumerate(obj.autocomplete_fields)
+            ]))
+
+    def _check_autocomplete_fields_item(self, obj, model, field_name, label):
+        """
+        Check that an item in `autocomplete_fields` is a ForeignKey or a
+        ManyToManyField and that the item has a related ModelAdmin with
+        search_fields defined.
+        """
+        try:
+            field = model._meta.get_field(field_name)
+        except FieldDoesNotExist:
+            return refer_to_missing_field(field=field_name, option=label, model=model, obj=obj, id='admin.E037')
+        else:
+            if not (field.many_to_many or field.many_to_one):
+                return must_be(
+                    'a foreign key or a many-to-many field',
+                    option=label, obj=obj, id='admin.E038'
+                )
+            related_admin = obj.admin_site._registry.get(field.remote_field.model)
+            if related_admin is None:
+                return [
+                    checks.Error(
+                        'An admin for model "%s" has to be registered '
+                        'to be referenced by %s.autocomplete_fields.' % (
+                            field.remote_field.model.__name__,
+                            type(obj).__name__,
+                        ),
+                        obj=obj.__class__,
+                        id='admin.E039',
+                    )
+                ]
+            elif not related_admin.search_fields:
+                return [
+                    checks.Error(
+                        '%s must define "search_fields", because it\'s '
+                        'referenced by %s.autocomplete_fields.' % (
+                            related_admin.__class__.__name__,
+                            type(obj).__name__,
+                        ),
+                        obj=obj.__class__,
+                        id='admin.E040',
+                    )
+                ]
+            return []
+
     def _check_raw_id_fields(self, obj):
         """ Check that `raw_id_fields` only contains field names that are listed
         on the model. """
@@ -87,10 +143,10 @@ class BaseModelAdminChecks(object):
         if not isinstance(obj.raw_id_fields, (list, tuple)):
             return must_be('a list or tuple', option='raw_id_fields', obj=obj, id='admin.E001')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_raw_id_fields_item(obj, obj.model, field_name, 'raw_id_fields[%d]' % index)
                 for index, field_name in enumerate(obj.raw_id_fields)
-            ]))
+            ))
 
     def _check_raw_id_fields_item(self, obj, model, field_name, label):
         """ Check an item of `raw_id_fields`, i.e. check that field named
@@ -136,10 +192,10 @@ class BaseModelAdminChecks(object):
                 )
             ]
 
-        return list(chain(*[
+        return list(chain.from_iterable(
             self._check_field_spec(obj, obj.model, field_name, 'fields')
             for field_name in obj.fields
-        ]))
+        ))
 
     def _check_fieldsets(self, obj):
         """ Check that fieldsets is properly formatted and doesn't contain
@@ -150,10 +206,10 @@ class BaseModelAdminChecks(object):
         elif not isinstance(obj.fieldsets, (list, tuple)):
             return must_be('a list or tuple', option='fieldsets', obj=obj, id='admin.E007')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_fieldsets_item(obj, obj.model, fieldset, 'fieldsets[%d]' % index)
                 for index, fieldset in enumerate(obj.fieldsets)
-            ]))
+            ))
 
     def _check_fieldsets_item(self, obj, model, fieldset, label):
         """ Check an item of `fieldsets`, i.e. check that this is a pair of a
@@ -185,10 +241,10 @@ class BaseModelAdminChecks(object):
                     id='admin.E012',
                 )
             ]
-        return list(chain(*[
+        return list(chain.from_iterable(
             self._check_field_spec(obj, model, fieldset_fields, '%s[1]["fields"]' % label)
             for fieldset_fields in fieldset[1]['fields']
-        ]))
+        ))
 
     def _check_field_spec(self, obj, model, fields, label):
         """ `fields` should be an item of `fields` or an item of
@@ -196,10 +252,10 @@ class BaseModelAdminChecks(object):
         field name or a tuple of field names. """
 
         if isinstance(fields, tuple):
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_field_spec_item(obj, model, field_name, "%s[%d]" % (label, index))
                 for index, field_name in enumerate(fields)
-            ]))
+            ))
         else:
             return self._check_field_spec_item(obj, model, fields, label)
 
@@ -217,10 +273,11 @@ class BaseModelAdminChecks(object):
                 # be an extra field on the form.
                 return []
             else:
-                if field.many_to_many and not field.remote_field.through._meta.auto_created:
+                if (isinstance(field, models.ManyToManyField) and
+                        not field.remote_field.through._meta.auto_created):
                     return [
                         checks.Error(
-                            "The value of '%s' cannot include the many-to-many field '%s' "
+                            "The value of '%s' cannot include the ManyToManyField '%s', "
                             "because that field manually specifies a relationship model."
                             % (label, field_name),
                             obj=obj.__class__,
@@ -250,8 +307,7 @@ class BaseModelAdminChecks(object):
 
     def _check_form(self, obj):
         """ Check that form subclasses BaseModelForm. """
-
-        if hasattr(obj, 'form') and not issubclass(obj.form, BaseModelForm):
+        if not issubclass(obj.form, BaseModelForm):
             return must_inherit_from(parent='BaseModelForm', option='form',
                                      obj=obj, id='admin.E016')
         else:
@@ -259,29 +315,23 @@ class BaseModelAdminChecks(object):
 
     def _check_filter_vertical(self, obj):
         """ Check that filter_vertical is a sequence of field names. """
-
-        if not hasattr(obj, 'filter_vertical'):
-            return []
-        elif not isinstance(obj.filter_vertical, (list, tuple)):
+        if not isinstance(obj.filter_vertical, (list, tuple)):
             return must_be('a list or tuple', option='filter_vertical', obj=obj, id='admin.E017')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_filter_item(obj, obj.model, field_name, "filter_vertical[%d]" % index)
                 for index, field_name in enumerate(obj.filter_vertical)
-            ]))
+            ))
 
     def _check_filter_horizontal(self, obj):
         """ Check that filter_horizontal is a sequence of field names. """
-
-        if not hasattr(obj, 'filter_horizontal'):
-            return []
-        elif not isinstance(obj.filter_horizontal, (list, tuple)):
+        if not isinstance(obj.filter_horizontal, (list, tuple)):
             return must_be('a list or tuple', option='filter_horizontal', obj=obj, id='admin.E018')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_filter_item(obj, obj.model, field_name, "filter_horizontal[%d]" % index)
                 for index, field_name in enumerate(obj.filter_horizontal)
-            ]))
+            ))
 
     def _check_filter_item(self, obj, model, field_name, label):
         """ Check one item of `filter_vertical` or `filter_horizontal`, i.e.
@@ -300,17 +350,14 @@ class BaseModelAdminChecks(object):
 
     def _check_radio_fields(self, obj):
         """ Check that `radio_fields` is a dictionary. """
-
-        if not hasattr(obj, 'radio_fields'):
-            return []
-        elif not isinstance(obj.radio_fields, dict):
+        if not isinstance(obj.radio_fields, dict):
             return must_be('a dictionary', option='radio_fields', obj=obj, id='admin.E021')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_radio_fields_key(obj, obj.model, field_name, 'radio_fields') +
                 self._check_radio_fields_value(obj, val, 'radio_fields["%s"]' % field_name)
                 for field_name, val in obj.radio_fields.items()
-            ]))
+            ))
 
     def _check_radio_fields_key(self, obj, model, field_name, label):
         """ Check that a key of `radio_fields` dictionary is name of existing
@@ -353,34 +400,28 @@ class BaseModelAdminChecks(object):
             return []
 
     def _check_view_on_site_url(self, obj):
-        if hasattr(obj, 'view_on_site'):
-            if not callable(obj.view_on_site) and not isinstance(obj.view_on_site, bool):
-                return [
-                    checks.Error(
-                        "The value of 'view_on_site' must be a callable or a boolean value.",
-                        obj=obj.__class__,
-                        id='admin.E025',
-                    )
-                ]
-            else:
-                return []
+        if not callable(obj.view_on_site) and not isinstance(obj.view_on_site, bool):
+            return [
+                checks.Error(
+                    "The value of 'view_on_site' must be a callable or a boolean value.",
+                    obj=obj.__class__,
+                    id='admin.E025',
+                )
+            ]
         else:
             return []
 
     def _check_prepopulated_fields(self, obj):
         """ Check that `prepopulated_fields` is a dictionary containing allowed
         field types. """
-
-        if not hasattr(obj, 'prepopulated_fields'):
-            return []
-        elif not isinstance(obj.prepopulated_fields, dict):
+        if not isinstance(obj.prepopulated_fields, dict):
             return must_be('a dictionary', option='prepopulated_fields', obj=obj, id='admin.E026')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_prepopulated_fields_key(obj, obj.model, field_name, 'prepopulated_fields') +
                 self._check_prepopulated_fields_value(obj, obj.model, val, 'prepopulated_fields["%s"]' % field_name)
                 for field_name, val in obj.prepopulated_fields.items()
-            ]))
+            ))
 
     def _check_prepopulated_fields_key(self, obj, model, field_name, label):
         """ Check a key of `prepopulated_fields` dictionary, i.e. check that it
@@ -393,11 +434,11 @@ class BaseModelAdminChecks(object):
             return refer_to_missing_field(field=field_name, option=label,
                                           model=model, obj=obj, id='admin.E027')
         else:
-            if field.many_to_many or isinstance(field, (models.DateTimeField, models.ForeignKey)):
+            if isinstance(field, (models.DateTimeField, models.ForeignKey, models.ManyToManyField)):
                 return [
                     checks.Error(
                         "The value of '%s' refers to '%s', which must not be a DateTimeField, "
-                        "a foreign key, or a many-to-many field." % (label, field_name),
+                        "a ForeignKey, a OneToOneField, or a ManyToManyField." % (label, field_name),
                         obj=obj.__class__,
                         id='admin.E028',
                     )
@@ -412,10 +453,10 @@ class BaseModelAdminChecks(object):
         if not isinstance(val, (list, tuple)):
             return must_be('a list or tuple', option=label, obj=obj, id='admin.E029')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_prepopulated_fields_value_item(obj, model, subfield_name, "%s[%r]" % (label, index))
                 for index, subfield_name in enumerate(val)
-            ]))
+            ))
 
     def _check_prepopulated_fields_value_item(self, obj, model, field_name, label):
         """ For `prepopulated_fields` equal to {"slug": ("title",)},
@@ -437,10 +478,10 @@ class BaseModelAdminChecks(object):
         elif not isinstance(obj.ordering, (list, tuple)):
             return must_be('a list or tuple', option='ordering', obj=obj, id='admin.E031')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_ordering_item(obj, obj.model, field_name, 'ordering[%d]' % index)
                 for index, field_name in enumerate(obj.ordering)
-            ]))
+            ))
 
     def _check_ordering_item(self, obj, model, field_name, label):
         """ Check that `ordering` refers to existing fields. """
@@ -457,14 +498,15 @@ class BaseModelAdminChecks(object):
             ]
         elif field_name == '?':
             return []
-        elif '__' in field_name:
+        elif LOOKUP_SEP in field_name:
             # Skip ordering in the format field1__field2 (FIXME: checking
             # this format would be nice, but it's a little fiddly).
             return []
         else:
             if field_name.startswith('-'):
                 field_name = field_name[1:]
-
+            if field_name == 'pk':
+                return []
             try:
                 model._meta.get_field(field_name)
             except FieldDoesNotExist:
@@ -480,10 +522,10 @@ class BaseModelAdminChecks(object):
         elif not isinstance(obj.readonly_fields, (list, tuple)):
             return must_be('a list or tuple', option='readonly_fields', obj=obj, id='admin.E034')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_readonly_fields_item(obj, obj.model, field_name, "readonly_fields[%d]" % index)
                 for index, field_name in enumerate(obj.readonly_fields)
-            ]))
+            ))
 
     def _check_readonly_fields_item(self, obj, model, field_name, label):
         if callable(field_name):
@@ -512,7 +554,7 @@ class BaseModelAdminChecks(object):
 class ModelAdminChecks(BaseModelAdminChecks):
 
     def check(self, admin_obj, **kwargs):
-        errors = super(ModelAdminChecks, self).check(admin_obj)
+        errors = super().check(admin_obj)
         errors.extend(self._check_save_as(admin_obj))
         errors.extend(self._check_save_on_top(admin_obj))
         errors.extend(self._check_inlines(admin_obj))
@@ -551,21 +593,21 @@ class ModelAdminChecks(BaseModelAdminChecks):
         if not isinstance(obj.inlines, (list, tuple)):
             return must_be('a list or tuple', option='inlines', obj=obj, id='admin.E103')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_inlines_item(obj, obj.model, item, "inlines[%d]" % index)
                 for index, item in enumerate(obj.inlines)
-            ]))
+            ))
 
     def _check_inlines_item(self, obj, model, inline, label):
         """ Check one inline model admin. """
-        inline_label = '.'.join([inline.__module__, inline.__name__])
+        inline_label = inline.__module__ + '.' + inline.__name__
 
-        from django.contrib.admin.options import BaseModelAdmin
+        from django.contrib.admin.options import InlineModelAdmin
 
-        if not issubclass(inline, BaseModelAdmin):
+        if not issubclass(inline, InlineModelAdmin):
             return [
                 checks.Error(
-                    "'%s' must inherit from 'BaseModelAdmin'." % inline_label,
+                    "'%s' must inherit from 'InlineModelAdmin'." % inline_label,
                     obj=obj.__class__,
                     id='admin.E104',
                 )
@@ -590,10 +632,10 @@ class ModelAdminChecks(BaseModelAdminChecks):
         if not isinstance(obj.list_display, (list, tuple)):
             return must_be('a list or tuple', option='list_display', obj=obj, id='admin.E107')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_list_display_item(obj, obj.model, item, "list_display[%d]" % index)
                 for index, item in enumerate(obj.list_display)
-            ]))
+            ))
 
     def _check_list_display_item(self, obj, model, item, label):
         if callable(item):
@@ -621,10 +663,10 @@ class ModelAdminChecks(BaseModelAdminChecks):
                         id='admin.E108',
                     )
                 ]
-            elif getattr(field, 'many_to_many', False):
+            elif isinstance(field, models.ManyToManyField):
                 return [
                     checks.Error(
-                        "The value of '%s' must not be a many-to-many field." % label,
+                        "The value of '%s' must not be a ManyToManyField." % label,
                         obj=obj.__class__,
                         id='admin.E109',
                     )
@@ -653,16 +695,19 @@ class ModelAdminChecks(BaseModelAdminChecks):
     def _check_list_display_links(self, obj):
         """ Check that list_display_links is a unique subset of list_display.
         """
+        from django.contrib.admin.options import ModelAdmin
 
         if obj.list_display_links is None:
             return []
         elif not isinstance(obj.list_display_links, (list, tuple)):
             return must_be('a list, a tuple, or None', option='list_display_links', obj=obj, id='admin.E110')
-        else:
-            return list(chain(*[
+        # Check only if ModelAdmin.get_list_display() isn't overridden.
+        elif obj.get_list_display.__func__ is ModelAdmin.get_list_display:
+            return list(chain.from_iterable(
                 self._check_list_display_links_item(obj, field_name, "list_display_links[%d]" % index)
                 for index, field_name in enumerate(obj.list_display_links)
-            ]))
+            ))
+        return []
 
     def _check_list_display_links_item(self, obj, field_name, label):
         if field_name not in obj.list_display:
@@ -682,10 +727,10 @@ class ModelAdminChecks(BaseModelAdminChecks):
         if not isinstance(obj.list_filter, (list, tuple)):
             return must_be('a list or tuple', option='list_filter', obj=obj, id='admin.E112')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_list_filter_item(obj, obj.model, item, "list_filter[%d]" % index)
                 for index, item in enumerate(obj.list_filter)
-            ]))
+            ))
 
     def _check_list_filter_item(self, obj, model, item, label):
         """
@@ -770,10 +815,10 @@ class ModelAdminChecks(BaseModelAdminChecks):
         if not isinstance(obj.list_editable, (list, tuple)):
             return must_be('a list or tuple', option='list_editable', obj=obj, id='admin.E120')
         else:
-            return list(chain(*[
+            return list(chain.from_iterable(
                 self._check_list_editable_item(obj, obj.model, item, "list_editable[%d]" % index)
                 for index, item in enumerate(obj.list_editable)
-            ]))
+            ))
 
     def _check_list_editable_item(self, obj, model, field_name, label):
         try:
@@ -840,12 +885,16 @@ class ModelAdminChecks(BaseModelAdminChecks):
             return []
         else:
             try:
-                field = obj.model._meta.get_field(obj.date_hierarchy)
-            except FieldDoesNotExist:
-                return refer_to_missing_field(
-                    option='date_hierarchy', field=obj.date_hierarchy,
-                    model=obj.model, obj=obj, id='admin.E127',
-                )
+                field = get_fields_from_path(obj.model, obj.date_hierarchy)[-1]
+            except (NotRelationField, FieldDoesNotExist):
+                return [
+                    checks.Error(
+                        "The value of 'date_hierarchy' refers to '%s', which "
+                        "does not refer to a Field." % obj.date_hierarchy,
+                        obj=obj.__class__,
+                        id='admin.E127',
+                    )
+                ]
             else:
                 if not isinstance(field, (models.DateField, models.DateTimeField)):
                     return must_be('a DateField or DateTimeField', option='date_hierarchy', obj=obj, id='admin.E128')
@@ -856,7 +905,7 @@ class ModelAdminChecks(BaseModelAdminChecks):
 class InlineModelAdminChecks(BaseModelAdminChecks):
 
     def check(self, inline_obj, **kwargs):
-        errors = super(InlineModelAdminChecks, self).check(inline_obj)
+        errors = super().check(inline_obj)
         parent_model = inline_obj.parent_model
         errors.extend(self._check_relation(inline_obj, parent_model))
         errors.extend(self._check_exclude_of_parent_model(inline_obj, parent_model))
@@ -869,7 +918,7 @@ class InlineModelAdminChecks(BaseModelAdminChecks):
     def _check_exclude_of_parent_model(self, obj, parent_model):
         # Do not perform more specific checks if the base checks result in an
         # error.
-        errors = super(InlineModelAdminChecks, self)._check_exclude(obj)
+        errors = super()._check_exclude(obj)
         if errors:
             return []
 
