@@ -2,37 +2,32 @@ import hashlib
 import logging
 from datetime import datetime
 
-from django.db.backends.utils import strip_quotes
+from django.db.backends.ddl_references import (
+    Columns, ForeignKeyName, IndexName, Statement, Table,
+)
+from django.db.backends.utils import split_identifier
 from django.db.models import Index
 from django.db.transaction import TransactionManagementError, atomic
-from django.utils import six, timezone
+from django.utils import timezone
 from django.utils.encoding import force_bytes
 
 logger = logging.getLogger('django.db.backends.schema')
 
 
 def _related_non_m2m_objects(old_field, new_field):
-    # Filters out m2m objects from reverse relations.
-    # Returns (old_relation, new_relation) tuples.
+    # Filter out m2m objects from reverse relations.
+    # Return (old_relation, new_relation) tuples.
     return zip(
         (obj for obj in old_field.model._meta.related_objects if not obj.field.many_to_many),
         (obj for obj in new_field.model._meta.related_objects if not obj.field.many_to_many)
     )
 
 
-class BaseDatabaseSchemaEditor(object):
+class BaseDatabaseSchemaEditor:
     """
-    This class (and its subclasses) are responsible for emitting schema-changing
+    This class and its subclasses are responsible for emitting schema-changing
     statements to the databases - model creation/removal/alteration, field
     renaming, index fiddling, and so on.
-
-    It is intended to eventually completely replace DatabaseCreation.
-
-    This class should be used by creating an instance for each set of schema
-    changes (e.g. a migration file), and by first calling start(),
-    then the relevant actions, and then commit(). This is necessary to allow
-    things like circular foreign key references - FKs will only be created once
-    commit() is called.
     """
 
     # Overrideable SQL templates
@@ -71,6 +66,8 @@ class BaseDatabaseSchemaEditor(object):
     sql_create_pk = "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s PRIMARY KEY (%(columns)s)"
     sql_delete_pk = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s"
 
+    sql_delete_procedure = 'DROP PROCEDURE %(procedure)s'
+
     def __init__(self, connection, collect_sql=False, atomic=True):
         self.connection = connection
         self.collect_sql = collect_sql
@@ -97,9 +94,7 @@ class BaseDatabaseSchemaEditor(object):
     # Core utility functions
 
     def execute(self, sql, params=()):
-        """
-        Executes the given SQL statement, with optional parameters.
-        """
+        """Execute the given SQL statement, with optional parameters."""
         # Don't perform the transactional DDL check if SQL is being collected
         # as it's not going to be executed anyway.
         if not self.collect_sql and self.connection.in_atomic_block and not self.connection.features.can_rollback_ddl:
@@ -107,6 +102,8 @@ class BaseDatabaseSchemaEditor(object):
                 "Executing DDL statements while in a transaction on databases "
                 "that can't perform a rollback is prohibited."
             )
+        # Account for non-string statement objects.
+        sql = str(sql)
         # Log the command we're running, then run it
         logger.debug("%s; (params %r)", sql, params, extra={'params': params, 'sql': sql})
         if self.collect_sql:
@@ -125,7 +122,7 @@ class BaseDatabaseSchemaEditor(object):
     @classmethod
     def _digest(cls, *args):
         """
-        Generates a 32-bit digest of a set of arguments that can be used to
+        Generate a 32-bit digest of a set of arguments that can be used to
         shorten identifying names.
         """
         h = hashlib.md5()
@@ -137,8 +134,8 @@ class BaseDatabaseSchemaEditor(object):
 
     def column_sql(self, model, field, include_default=False):
         """
-        Takes a field and returns its column definition.
-        The field must already have had set_attributes_from_name called.
+        Take a field and return its column definition.
+        The field must already have had set_attributes_from_name() called.
         """
         # Get the column's type and use that as the basis of the SQL
         db_params = field.db_parameters(connection=self.connection)
@@ -200,16 +197,14 @@ class BaseDatabaseSchemaEditor(object):
         )
 
     def effective_default(self, field):
-        """
-        Returns a field's effective database default value
-        """
+        """Return a field's effective database default value."""
         if field.has_default():
             default = field.get_default()
         elif not field.null and field.blank and field.empty_strings_allowed:
             if field.get_internal_type() == "BinaryField":
-                default = six.binary_type()
+                default = bytes()
             else:
-                default = six.text_type()
+                default = str()
         elif getattr(field, 'auto_now', False) or getattr(field, 'auto_now_add', False):
             default = datetime.now()
             internal_type = field.get_internal_type()
@@ -231,7 +226,7 @@ class BaseDatabaseSchemaEditor(object):
 
     def quote_value(self, value):
         """
-        Returns a quoted version of the value so it's safe to use in an SQL
+        Return a quoted version of the value so it's safe to use in an SQL
         string. This is not safe against injection from user code; it is
         intended only for use in making SQL scripts or preparing default values
         for particularly tricky backends (defaults are not user-defined, though,
@@ -243,8 +238,8 @@ class BaseDatabaseSchemaEditor(object):
 
     def create_model(self, model):
         """
-        Takes a model and creates a table for it in the database.
-        Will also create any accompanying indexes or unique constraints.
+        Create a table and any accompanying indexes or unique constraints for
+        the given `model`.
         """
         # Create column SQL, add FK deferreds if needed
         column_sqls = []
@@ -267,13 +262,13 @@ class BaseDatabaseSchemaEditor(object):
             if field.remote_field and field.db_constraint:
                 to_table = field.remote_field.model._meta.db_table
                 to_column = field.remote_field.model._meta.get_field(field.remote_field.field_name).column
-                if self.connection.features.supports_foreign_keys:
-                    self.deferred_sql.append(self._create_fk_sql(model, field, "_fk_%(to_table)s_%(to_column)s"))
-                elif self.sql_create_inline_fk:
+                if self.sql_create_inline_fk:
                     definition += " " + self.sql_create_inline_fk % {
                         "to_table": self.quote_name(to_table),
                         "to_column": self.quote_name(to_column),
                     }
+                elif self.connection.features.supports_foreign_keys:
+                    self.deferred_sql.append(self._create_fk_sql(model, field, "_fk_%(to_table)s_%(to_column)s"))
             # Add the SQL to our big list
             column_sqls.append("%s %s" % (
                 self.quote_name(field.column),
@@ -311,9 +306,7 @@ class BaseDatabaseSchemaEditor(object):
                 self.create_model(field.remote_field.through)
 
     def delete_model(self, model):
-        """
-        Deletes a model from the database.
-        """
+        """Delete a model from the database."""
         # Handle auto-created intermediary models
         for field in model._meta.local_many_to_many:
             if field.remote_field.through._meta.auto_created:
@@ -323,27 +316,27 @@ class BaseDatabaseSchemaEditor(object):
         self.execute(self.sql_delete_table % {
             "table": self.quote_name(model._meta.db_table),
         })
+        # Remove all deferred statements referencing the deleted table.
+        for sql in list(self.deferred_sql):
+            if isinstance(sql, Statement) and sql.references_table(model._meta.db_table):
+                self.deferred_sql.remove(sql)
 
     def add_index(self, model, index):
-        """
-        Add an index on a model.
-        """
+        """Add an index on a model."""
         self.execute(index.create_sql(model, self))
 
     def remove_index(self, model, index):
-        """
-        Remove an index from a model.
-        """
+        """Remove an index from a model."""
         self.execute(index.remove_sql(model, self))
 
     def alter_unique_together(self, model, old_unique_together, new_unique_together):
         """
-        Deals with a model changing its unique_together.
-        Note: The input unique_togethers must be doubly-nested, not the single-
-        nested ["foo", "bar"] format.
+        Deal with a model changing its unique_together. The input
+        unique_togethers must be doubly-nested, not the single-nested
+        ["foo", "bar"] format.
         """
-        olds = set(tuple(fields) for fields in old_unique_together)
-        news = set(tuple(fields) for fields in new_unique_together)
+        olds = {tuple(fields) for fields in old_unique_together}
+        news = {tuple(fields) for fields in new_unique_together}
         # Deleted uniques
         for fields in olds.difference(news):
             self._delete_composed_index(model, fields, {'unique': True}, self.sql_delete_unique)
@@ -354,12 +347,12 @@ class BaseDatabaseSchemaEditor(object):
 
     def alter_index_together(self, model, old_index_together, new_index_together):
         """
-        Deals with a model changing its index_together.
-        Note: The input index_togethers must be doubly-nested, not the single-
-        nested ["foo", "bar"] format.
+        Deal with a model changing its index_together. The input
+        index_togethers must be doubly-nested, not the single-nested
+        ["foo", "bar"] format.
         """
-        olds = set(tuple(fields) for fields in old_index_together)
-        news = set(tuple(fields) for fields in new_index_together)
+        olds = {tuple(fields) for fields in old_index_together}
+        news = {tuple(fields) for fields in new_index_together}
         # Deleted indexes
         for fields in olds.difference(news):
             self._delete_composed_index(model, fields, {'index': True}, self.sql_delete_index)
@@ -380,9 +373,7 @@ class BaseDatabaseSchemaEditor(object):
         self.execute(self._delete_constraint_sql(sql, model, constraint_names[0]))
 
     def alter_db_table(self, model, old_db_table, new_db_table):
-        """
-        Renames the table a model points to.
-        """
+        """Rename the table a model points to."""
         if (old_db_table == new_db_table or
             (self.connection.features.ignores_table_name_case and
                 old_db_table.lower() == new_db_table.lower())):
@@ -391,11 +382,13 @@ class BaseDatabaseSchemaEditor(object):
             "old_table": self.quote_name(old_db_table),
             "new_table": self.quote_name(new_db_table),
         })
+        # Rename all references to the old table name.
+        for sql in self.deferred_sql:
+            if isinstance(sql, Statement):
+                sql.rename_table_references(old_db_table, new_db_table)
 
     def alter_db_tablespace(self, model, old_db_tablespace, new_db_tablespace):
-        """
-        Moves a model's table between tablespaces
-        """
+        """Move a model's table between tablespaces."""
         self.execute(self.sql_retablespace_table % {
             "table": self.quote_name(model._meta.db_table),
             "old_tablespace": self.quote_name(old_db_tablespace),
@@ -404,9 +397,8 @@ class BaseDatabaseSchemaEditor(object):
 
     def add_field(self, model, field):
         """
-        Creates a field on a model.
-        Usually involves adding a column, but may involve adding a
-        table instead (for M2M fields)
+        Create a field on a model. Usually involves adding a column, but may
+        involve adding a table instead (for M2M fields).
         """
         # Special-case implicit M2M tables
         if field.many_to_many and field.remote_field.through._meta.auto_created:
@@ -430,14 +422,12 @@ class BaseDatabaseSchemaEditor(object):
         # Drop the default if we need to
         # (Django usually does not use in-database defaults)
         if not self.skip_default(field) and self.effective_default(field) is not None:
+            changes_sql, params = self._alter_column_default_sql(model, None, field, drop=True)
             sql = self.sql_alter_column % {
                 "table": self.quote_name(model._meta.db_table),
-                "changes": self.sql_alter_column_no_default % {
-                    "column": self.quote_name(field.column),
-                    "type": db_params['type'],
-                }
+                "changes": changes_sql,
             }
-            self.execute(sql)
+            self.execute(sql, params)
         # Add an index, if required
         self.deferred_sql.extend(self._field_indexes_sql(model, field))
         # Add any FK constraints later
@@ -449,7 +439,7 @@ class BaseDatabaseSchemaEditor(object):
 
     def remove_field(self, model, field):
         """
-        Removes a field from a model. Usually involves deleting a column,
+        Remove a field from a model. Usually involves deleting a column,
         but for M2Ms may involve deleting a table.
         """
         # Special-case implicit M2M tables
@@ -472,14 +462,18 @@ class BaseDatabaseSchemaEditor(object):
         # Reset connection if required
         if self.connection.features.connection_persists_old_columns:
             self.connection.close()
+        # Remove all deferred statements referencing the deleted table.
+        for sql in list(self.deferred_sql):
+            if isinstance(sql, Statement) and sql.references_column(model._meta.db_table, field.column):
+                self.deferred_sql.remove(sql)
 
     def alter_field(self, model, old_field, new_field, strict=False):
         """
-        Allows a field's type, uniqueness, nullability, default, column,
-        constraints etc. to be modified.
-        Requires a copy of the old field as well so we can only perform
-        changes that are required.
-        If strict is true, raises errors if the old column does not match old_field precisely.
+        Allow a field's type, uniqueness, nullability, default, column,
+        constraints, etc. to be modified.
+        `old_field` is required to compute the necessary changes.
+        If `strict` is True, raise errors if the old column does not match
+        `old_field` precisely.
         """
         # Ensure this field is even column-based
         old_db_params = old_field.db_parameters(connection=self.connection)
@@ -516,8 +510,7 @@ class BaseDatabaseSchemaEditor(object):
 
     def _alter_field(self, model, old_field, new_field, old_type, new_type,
                      old_db_params, new_db_params, strict=False):
-        """Actually perform a "physical" (non-ManyToMany) field update."""
-
+        """Perform a "physical" (non-ManyToMany) field update."""
         # Drop any FK constraints, we'll remake them later
         fks_dropped = set()
         if old_field.remote_field and old_field.db_constraint:
@@ -543,9 +536,15 @@ class BaseDatabaseSchemaEditor(object):
                 ))
             for constraint_name in constraint_names:
                 self.execute(self._delete_constraint_sql(self.sql_delete_unique, model, constraint_name))
-        # Drop incoming FK constraints if we're a primary key and things are going
-        # to change.
-        if old_field.primary_key and new_field.primary_key and old_type != new_type:
+        # Drop incoming FK constraints if the field is a primary key or unique,
+        # which might be a to_field target, and things are going to change.
+        drop_foreign_keys = (
+            (
+                (old_field.primary_key and new_field.primary_key) or
+                (old_field.unique and new_field.unique)
+            ) and old_type != new_type
+        )
+        if drop_foreign_keys:
             # '_meta.related_field' also contains M2M reverse fields, these
             # will be filtered out
             for _old_rel, new_rel in _related_non_m2m_objects(old_field, new_field):
@@ -591,15 +590,17 @@ class BaseDatabaseSchemaEditor(object):
         # Have they renamed the column?
         if old_field.column != new_field.column:
             self.execute(self._rename_field_sql(model._meta.db_table, old_field, new_field, new_type))
+            # Rename all references to the renamed column.
+            for sql in self.deferred_sql:
+                if isinstance(sql, Statement):
+                    sql.rename_column_references(model._meta.db_table, old_field.column, new_field.column)
         # Next, start accumulating actions to do
         actions = []
         null_actions = []
         post_actions = []
         # Type change?
         if old_type != new_type:
-            fragment, other_actions = self._alter_column_type_sql(
-                model._meta.db_table, old_field, new_field, new_type
-            )
+            fragment, other_actions = self._alter_column_type_sql(model, old_field, new_field, new_type)
             actions.append(fragment)
             post_actions.extend(other_actions)
         # When changing a column NULL constraint to NOT NULL with a given
@@ -612,54 +613,19 @@ class BaseDatabaseSchemaEditor(object):
         old_default = self.effective_default(old_field)
         new_default = self.effective_default(new_field)
         needs_database_default = (
+            old_field.null and
+            not new_field.null and
             old_default != new_default and
             new_default is not None and
             not self.skip_default(new_field)
         )
         if needs_database_default:
-            if self.connection.features.requires_literal_defaults:
-                # Some databases can't take defaults as a parameter (oracle)
-                # If this is the case, the individual schema backend should
-                # implement prepare_default
-                actions.append((
-                    self.sql_alter_column_default % {
-                        "column": self.quote_name(new_field.column),
-                        "type": new_type,
-                        "default": self.prepare_default(new_default),
-                    },
-                    [],
-                ))
-            else:
-                actions.append((
-                    self.sql_alter_column_default % {
-                        "column": self.quote_name(new_field.column),
-                        "type": new_type,
-                        "default": "%s",
-                    },
-                    [new_default],
-                ))
+            actions.append(self._alter_column_default_sql(model, old_field, new_field))
         # Nullability change?
         if old_field.null != new_field.null:
-            if (self.connection.features.interprets_empty_strings_as_nulls and
-                    new_field.get_internal_type() in ("CharField", "TextField")):
-                # The field is nullable in the database anyway, leave it alone
-                pass
-            elif new_field.null:
-                null_actions.append((
-                    self.sql_alter_column_null % {
-                        "column": self.quote_name(new_field.column),
-                        "type": new_type,
-                    },
-                    [],
-                ))
-            else:
-                null_actions.append((
-                    self.sql_alter_column_not_null % {
-                        "column": self.quote_name(new_field.column),
-                        "type": new_type,
-                    },
-                    [],
-                ))
+            fragment = self._alter_column_null_sql(model, old_field, new_field)
+            if fragment:
+                null_actions.append(fragment)
         # Only if we have a default and there is a change from NULL to NOT NULL
         four_way_default_alteration = (
             new_field.has_default() and
@@ -732,19 +698,14 @@ class BaseDatabaseSchemaEditor(object):
         # will always come along and replace it.
         if not old_field.primary_key and new_field.primary_key:
             # First, drop the old PK
-            constraint_names = self._constraint_names(model, primary_key=True)
-            if strict and len(constraint_names) != 1:
-                raise ValueError("Found wrong number (%s) of PK constraints for %s" % (
-                    len(constraint_names),
-                    model._meta.db_table,
-                ))
-            for constraint_name in constraint_names:
-                self.execute(self._delete_constraint_sql(self.sql_delete_pk, model, constraint_name))
+            self._delete_primary_key(model, strict)
             # Make the new one
             self.execute(
                 self.sql_create_pk % {
                     "table": self.quote_name(model._meta.db_table),
-                    "name": self.quote_name(self._create_index_name(model, [new_field.column], suffix="_pk")),
+                    "name": self.quote_name(
+                        self._create_index_name(model._meta.db_table, [new_field.column], suffix="_pk")
+                    ),
                     "columns": self.quote_name(new_field.column),
                 }
             )
@@ -755,7 +716,7 @@ class BaseDatabaseSchemaEditor(object):
             rel_db_params = new_rel.field.db_parameters(connection=self.connection)
             rel_type = rel_db_params['type']
             fragment, other_actions = self._alter_column_type_sql(
-                new_rel.related_model._meta.db_table, old_rel.field, new_rel.field, rel_type
+                new_rel.related_model, old_rel.field, new_rel.field, rel_type
             )
             self.execute(
                 self.sql_alter_column % {
@@ -772,7 +733,7 @@ class BaseDatabaseSchemaEditor(object):
                 new_field.db_constraint):
             self.execute(self._create_fk_sql(model, new_field, "_fk_%(to_table)s_%(to_column)s"))
         # Rebuild FKs that pointed to us if we previously had to drop them
-        if old_field.primary_key and new_field.primary_key and old_type != new_type:
+        if drop_foreign_keys:
             for rel in new_field.model._meta.related_objects:
                 if not rel.many_to_many and rel.field.db_constraint:
                     self.execute(self._create_fk_sql(rel.related_model, rel.field, "_fk"))
@@ -781,7 +742,9 @@ class BaseDatabaseSchemaEditor(object):
             self.execute(
                 self.sql_create_check % {
                     "table": self.quote_name(model._meta.db_table),
-                    "name": self.quote_name(self._create_index_name(model, [new_field.column], suffix="_check")),
+                    "name": self.quote_name(
+                        self._create_index_name(model._meta.db_table, [new_field.column], suffix="_check")
+                    ),
                     "column": self.quote_name(new_field.column),
                     "check": new_db_params['check'],
                 }
@@ -789,27 +752,78 @@ class BaseDatabaseSchemaEditor(object):
         # Drop the default if we need to
         # (Django usually does not use in-database defaults)
         if needs_database_default:
+            changes_sql, params = self._alter_column_default_sql(model, old_field, new_field, drop=True)
             sql = self.sql_alter_column % {
                 "table": self.quote_name(model._meta.db_table),
-                "changes": self.sql_alter_column_no_default % {
-                    "column": self.quote_name(new_field.column),
-                    "type": new_type,
-                }
+                "changes": changes_sql,
             }
-            self.execute(sql)
+            self.execute(sql, params)
         # Reset connection if required
         if self.connection.features.connection_persists_old_columns:
             self.connection.close()
 
-    def _alter_column_type_sql(self, table, old_field, new_field, new_type):
+    def _alter_column_null_sql(self, model, old_field, new_field):
+        """
+        Hook to specialize column null alteration.
+
+        Return a (sql, params) fragment to set a column to null or non-null
+        as required by new_field, or None if no changes are required.
+        """
+        if (self.connection.features.interprets_empty_strings_as_nulls and
+                new_field.get_internal_type() in ("CharField", "TextField")):
+            # The field is nullable in the database anyway, leave it alone.
+            return
+        else:
+            new_db_params = new_field.db_parameters(connection=self.connection)
+            sql = self.sql_alter_column_null if new_field.null else self.sql_alter_column_not_null
+            return (
+                sql % {
+                    'column': self.quote_name(new_field.column),
+                    'type': new_db_params['type'],
+                },
+                [],
+            )
+
+    def _alter_column_default_sql(self, model, old_field, new_field, drop=False):
+        """
+        Hook to specialize column default alteration.
+
+        Return a (sql, params) fragment to add or drop (depending on the drop
+        argument) a default to new_field's column.
+        """
+        new_default = self.effective_default(new_field)
+        default = '%s'
+        params = [new_default]
+
+        if drop:
+            params = []
+        elif self.connection.features.requires_literal_defaults:
+            # Some databases (Oracle) can't take defaults as a parameter
+            # If this is the case, the SchemaEditor for that database should
+            # implement prepare_default().
+            default = self.prepare_default(new_default)
+            params = []
+
+        new_db_params = new_field.db_parameters(connection=self.connection)
+        sql = self.sql_alter_column_no_default if drop else self.sql_alter_column_default
+        return (
+            sql % {
+                'column': self.quote_name(new_field.column),
+                'type': new_db_params['type'],
+                'default': default,
+            },
+            params,
+        )
+
+    def _alter_column_type_sql(self, model, old_field, new_field, new_type):
         """
         Hook to specialize column type alteration for different backends,
         for cases when a creation type is different to an alteration type
         (e.g. SERIAL in PostgreSQL, PostGIS fields).
 
-        Should return two things; an SQL fragment of (sql, params) to insert
-        into an ALTER TABLE statement, and a list of extra (sql, params) tuples
-        to run once the field is altered.
+        Return a two-tuple of: an SQL fragment of (sql, params) to insert into
+        an ALTER TABLE statement and a list of extra (sql, params) tuples to
+        run once the field is altered.
         """
         return (
             (
@@ -823,9 +837,7 @@ class BaseDatabaseSchemaEditor(object):
         )
 
     def _alter_many_to_many(self, model, old_field, new_field, strict):
-        """
-        Alters M2Ms to repoint their to= endpoints.
-        """
+        """Alter M2Ms to repoint their to= endpoints."""
         # Rename the through table
         if old_field.remote_field.through._meta.db_table != new_field.remote_field.through._meta.db_table:
             self.alter_db_table(old_field.remote_field.through, old_field.remote_field.through._meta.db_table,
@@ -845,14 +857,14 @@ class BaseDatabaseSchemaEditor(object):
             new_field.remote_field.through._meta.get_field(new_field.m2m_field_name()),
         )
 
-    def _create_index_name(self, model, column_names, suffix=""):
+    def _create_index_name(self, table_name, column_names, suffix=""):
         """
-        Generates a unique name for an index/unique constraint.
+        Generate a unique name for an index/unique constraint.
 
         The name is divided into 3 parts: the table name, the column names,
         and a unique digest and suffix.
         """
-        table_name = strip_quotes(model._meta.db_table)
+        _, table_name = split_identifier(table_name)
         hash_data = [table_name] + list(column_names)
         hash_suffix_part = '%s%s' % (self._digest(*hash_data), suffix)
         max_length = self.connection.ops.max_name_length() or 200
@@ -875,38 +887,47 @@ class BaseDatabaseSchemaEditor(object):
             index_name = "D%s" % index_name[:-1]
         return index_name
 
-    def _get_index_tablespace_sql(self, model, fields):
-        if len(fields) == 1 and fields[0].db_tablespace:
-            tablespace_sql = self.connection.ops.tablespace_sql(fields[0].db_tablespace)
-        elif model._meta.db_tablespace:
-            tablespace_sql = self.connection.ops.tablespace_sql(model._meta.db_tablespace)
-        else:
-            tablespace_sql = ""
-        if tablespace_sql:
-            tablespace_sql = " " + tablespace_sql
-        return tablespace_sql
+    def _get_index_tablespace_sql(self, model, fields, db_tablespace=None):
+        if db_tablespace is None:
+            if len(fields) == 1 and fields[0].db_tablespace:
+                db_tablespace = fields[0].db_tablespace
+            elif model._meta.db_tablespace:
+                db_tablespace = model._meta.db_tablespace
+        if db_tablespace is not None:
+            return ' ' + self.connection.ops.tablespace_sql(db_tablespace)
+        return ''
 
-    def _create_index_sql(self, model, fields, suffix="", sql=None):
+    def _create_index_sql(self, model, fields, *, name=None, suffix='', using='',
+                          db_tablespace=None, col_suffixes=(), sql=None):
         """
         Return the SQL statement to create the index for one or several fields.
         `sql` can be specified if the syntax differs from the standard (GIS
         indexes, ...).
         """
-        tablespace_sql = self._get_index_tablespace_sql(model, fields)
+        tablespace_sql = self._get_index_tablespace_sql(model, fields, db_tablespace=db_tablespace)
         columns = [field.column for field in fields]
         sql_create_index = sql or self.sql_create_index
-        return sql_create_index % {
-            "table": self.quote_name(model._meta.db_table),
-            "name": self.quote_name(self._create_index_name(model, columns, suffix=suffix)),
-            "using": "",
-            "columns": ", ".join(self.quote_name(column) for column in columns),
-            "extra": tablespace_sql,
-        }
+        table = model._meta.db_table
+
+        def create_index_name(*args, **kwargs):
+            nonlocal name
+            if name is None:
+                name = self._create_index_name(*args, **kwargs)
+            return self.quote_name(name)
+
+        return Statement(
+            sql_create_index,
+            table=Table(table, self.quote_name),
+            name=IndexName(table, columns, suffix, create_index_name),
+            using=using,
+            columns=Columns(table, columns, self.quote_name, col_suffixes=col_suffixes),
+            extra=tablespace_sql,
+        )
 
     def _model_indexes_sql(self, model):
         """
-        Return all index SQL statements (field indexes, index_together,
-        Meta.indexes) for the specified model, as a list.
+        Return a list of all index SQL statements (field indexes,
+        index_together, Meta.indexes) for the specified model.
         """
         if not model._meta.managed or model._meta.proxy or model._meta.swapped:
             return []
@@ -947,26 +968,28 @@ class BaseDatabaseSchemaEditor(object):
         from_column = field.column
         to_table = field.target_field.model._meta.db_table
         to_column = field.target_field.column
-        suffix = suffix % {
-            "to_table": to_table,
-            "to_column": to_column,
-        }
 
-        return self.sql_create_fk % {
-            "table": self.quote_name(from_table),
-            "name": self.quote_name(self._create_index_name(model, [from_column], suffix=suffix)),
-            "column": self.quote_name(from_column),
-            "to_table": self.quote_name(to_table),
-            "to_column": self.quote_name(to_column),
-            "deferrable": self.connection.ops.deferrable_sql(),
-        }
+        def create_fk_name(*args, **kwargs):
+            return self.quote_name(self._create_index_name(*args, **kwargs))
+
+        return Statement(
+            self.sql_create_fk,
+            table=Table(from_table, self.quote_name),
+            name=ForeignKeyName(from_table, [from_column], to_table, [to_column], suffix, create_fk_name),
+            column=Columns(from_table, [from_column], self.quote_name),
+            to_table=Table(to_table, self.quote_name),
+            to_column=Columns(to_table, [to_column], self.quote_name),
+            deferrable=self.connection.ops.deferrable_sql(),
+        )
 
     def _create_unique_sql(self, model, columns):
-        return self.sql_create_unique % {
-            "table": self.quote_name(model._meta.db_table),
-            "name": self.quote_name(self._create_index_name(model, columns, suffix="_uniq")),
-            "columns": ", ".join(self.quote_name(column) for column in columns),
-        }
+        table = model._meta.db_table
+        return Statement(
+            self.sql_create_unique,
+            table=Table(table, self.quote_name),
+            name=IndexName(table, columns, '_uniq', self._create_index_name),
+            columns=Columns(table, columns, self.quote_name),
+        )
 
     def _delete_constraint_sql(self, template, model, name):
         return template % {
@@ -977,9 +1000,7 @@ class BaseDatabaseSchemaEditor(object):
     def _constraint_names(self, model, column_names=None, unique=None,
                           primary_key=None, index=None, foreign_key=None,
                           check=None, type_=None):
-        """
-        Returns all constraint names matching the columns and conditions
-        """
+        """Return all constraint names matching the columns and conditions."""
         if column_names is not None:
             column_names = [
                 self.connection.introspection.column_name_converter(name)
@@ -1004,3 +1025,20 @@ class BaseDatabaseSchemaEditor(object):
                     continue
                 result.append(name)
         return result
+
+    def _delete_primary_key(self, model, strict=False):
+        constraint_names = self._constraint_names(model, primary_key=True)
+        if strict and len(constraint_names) != 1:
+            raise ValueError('Found wrong number (%s) of PK constraints for %s' % (
+                len(constraint_names),
+                model._meta.db_table,
+            ))
+        for constraint_name in constraint_names:
+            self.execute(self._delete_constraint_sql(self.sql_delete_pk, model, constraint_name))
+
+    def remove_procedure(self, procedure_name, param_types=()):
+        sql = self.sql_delete_procedure % {
+            'procedure': self.quote_name(procedure_name),
+            'param_types': ','.join(param_types),
+        }
+        self.execute(sql)
