@@ -5,32 +5,28 @@ from decimal import Decimal
 
 from django.apps.registry import Apps
 from django.db.backends.base.schema import BaseDatabaseSchemaEditor
-from django.utils import six
+from django.db.backends.ddl_references import Statement
+from django.db.transaction import atomic
+from django.db.utils import NotSupportedError
 
 
 class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
 
     sql_delete_table = "DROP TABLE %(table)s"
-    sql_create_inline_fk = "REFERENCES %(to_table)s (%(to_column)s)"
+    sql_create_fk = None
+    sql_create_inline_fk = "REFERENCES %(to_table)s (%(to_column)s) DEFERRABLE INITIALLY DEFERRED"
     sql_create_unique = "CREATE UNIQUE INDEX %(name)s ON %(table)s (%(columns)s)"
     sql_delete_unique = "DROP INDEX %(name)s"
 
     def __enter__(self):
-        with self.connection.cursor() as c:
-            # Some SQLite schema alterations need foreign key constraints to be
-            # disabled. This is the default in SQLite but can be changed with a
-            # build flag and might change in future, so can't be relied upon.
-            # We enforce it here for the duration of the transaction.
-            c.execute('PRAGMA foreign_keys')
-            self._initial_pragma_fk = c.fetchone()[0]
-            c.execute('PRAGMA foreign_keys = 0')
-        return super(DatabaseSchemaEditor, self).__enter__()
+        # Some SQLite schema alterations need foreign key constraints to be
+        # disabled. Enforce it here for the duration of the transaction.
+        self.connection.disable_constraint_checking()
+        return super().__enter__()
 
     def __exit__(self, exc_type, exc_value, traceback):
-        super(DatabaseSchemaEditor, self).__exit__(exc_type, exc_value, traceback)
-        with self.connection.cursor() as c:
-            # Restore initial FK setting - PRAGMA values can't be parametrized
-            c.execute('PRAGMA foreign_keys = %s' % int(self._initial_pragma_fk))
+        super().__exit__(exc_type, exc_value, traceback)
+        self.connection.enable_constraint_checking()
 
     def quote_value(self, value):
         # The backend "mostly works" without this function and there are use
@@ -46,15 +42,13 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
         # Manual emulation of SQLite parameter quoting
         if isinstance(value, type(True)):
             return str(int(value))
-        elif isinstance(value, (Decimal, float)):
+        elif isinstance(value, (Decimal, float, int)):
             return str(value)
-        elif isinstance(value, six.integer_types):
-            return str(value)
-        elif isinstance(value, six.string_types):
-            return "'%s'" % six.text_type(value).replace("\'", "\'\'")
+        elif isinstance(value, str):
+            return "'%s'" % value.replace("\'", "\'\'")
         elif value is None:
             return "NULL"
-        elif isinstance(value, (bytes, bytearray, six.memoryview)):
+        elif isinstance(value, (bytes, bytearray, memoryview)):
             # Bytes are only allowed for BLOB fields, encoded as string
             # literals containing hexadecimal data and preceded by a single "X"
             # character:
@@ -67,8 +61,59 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
         else:
             raise ValueError("Cannot quote parameter value %r of type %s" % (value, type(value)))
 
-    def _remake_table(self, model, create_fields=[], delete_fields=[], alter_fields=[], override_uniques=None,
-                      override_indexes=None):
+    def alter_db_table(self, model, old_db_table, new_db_table, disable_constraints=True):
+        if model._meta.related_objects and disable_constraints:
+            if self.connection.in_atomic_block:
+                raise NotSupportedError((
+                    'Renaming the %r table while in a transaction is not '
+                    'supported on SQLite because it would break referential '
+                    'integrity. Try adding `atomic = False` to the Migration class.'
+                ) % old_db_table)
+            self.connection.enable_constraint_checking()
+            super().alter_db_table(model, old_db_table, new_db_table)
+            self.connection.disable_constraint_checking()
+        else:
+            super().alter_db_table(model, old_db_table, new_db_table)
+
+    def alter_field(self, model, old_field, new_field, strict=False):
+        old_field_name = old_field.name
+        if (new_field.name != old_field_name and
+                any(r.field_name == old_field.name for r in model._meta.related_objects)):
+            if self.connection.in_atomic_block:
+                raise NotSupportedError((
+                    'Renaming the %r.%r column while in a transaction is not '
+                    'supported on SQLite because it would break referential '
+                    'integrity. Try adding `atomic = False` to the Migration class.'
+                ) % (model._meta.db_table, old_field_name))
+            with atomic(self.connection.alias):
+                super().alter_field(model, old_field, new_field, strict=strict)
+                # Follow SQLite's documented procedure for performing changes
+                # that don't affect the on-disk content.
+                # https://sqlite.org/lang_altertable.html#otheralter
+                with self.connection.cursor() as cursor:
+                    schema_version = cursor.execute('PRAGMA schema_version').fetchone()[0]
+                    cursor.execute('PRAGMA writable_schema = 1')
+                    table_name = model._meta.db_table
+                    references_template = ' REFERENCES "%s" ("%%s") ' % table_name
+                    old_column_name = old_field.get_attname_column()[1]
+                    new_column_name = new_field.get_attname_column()[1]
+                    search = references_template % old_column_name
+                    replacement = references_template % new_column_name
+                    cursor.execute('UPDATE sqlite_master SET sql = replace(sql, %s, %s)', (search, replacement))
+                    cursor.execute('PRAGMA schema_version = %d' % (schema_version + 1))
+                    cursor.execute('PRAGMA writable_schema = 0')
+                    # The integrity check will raise an exception and rollback
+                    # the transaction if the sqlite_master updates corrupt the
+                    # database.
+                    cursor.execute('PRAGMA integrity_check')
+            # Perform a VACUUM to refresh the database representation from
+            # the sqlite_master table.
+            with self.connection.cursor() as cursor:
+                cursor.execute('VACUUM')
+        else:
+            super().alter_field(model, old_field, new_field, strict=strict)
+
+    def _remake_table(self, model, create_field=None, delete_field=None, alter_field=None):
         """
         Shortcut to transform a model from old_model into new_model
 
@@ -96,7 +141,8 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
         # If any of the new or altered fields is introducing a new PK,
         # remove the old one
         restore_pk_field = None
-        if any(f.primary_key for f in create_fields) or any(n.primary_key for o, n in alter_fields):
+        if getattr(create_field, 'primary_key', False) or (
+                alter_field and getattr(alter_field[1], 'primary_key', False)):
             for name, field in list(body.items()):
                 if field.primary_key:
                     field.primary_key = False
@@ -105,15 +151,16 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
                         del body[name]
                         del mapping[field.column]
         # Add in any created fields
-        for field in create_fields:
-            body[field.name] = field
+        if create_field:
+            body[create_field.name] = create_field
             # Choose a default and insert it into the copy map
-            if not field.many_to_many and field.concrete:
-                mapping[field.column] = self.quote_value(
-                    self.effective_default(field)
+            if not create_field.many_to_many and create_field.concrete:
+                mapping[create_field.column] = self.quote_value(
+                    self.effective_default(create_field)
                 )
         # Add in any altered fields
-        for (old_field, new_field) in alter_fields:
+        if alter_field:
+            old_field, new_field = alter_field
             body.pop(old_field.name, None)
             mapping.pop(old_field.column, None)
             body[new_field.name] = new_field
@@ -127,12 +174,12 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
                 mapping[new_field.column] = self.quote_name(old_field.column)
             rename_mapping[old_field.name] = new_field.name
         # Remove any deleted fields
-        for field in delete_fields:
-            del body[field.name]
-            del mapping[field.column]
+        if delete_field:
+            del body[delete_field.name]
+            del mapping[delete_field.column]
             # Remove any implicit M2M tables
-            if field.many_to_many and field.remote_field.through._meta.auto_created:
-                return self.delete_model(field.remote_field.through)
+            if delete_field.many_to_many and delete_field.remote_field.through._meta.auto_created:
+                return self.delete_model(delete_field.remote_field.through)
         # Work inside a new app registry
         apps = Apps()
 
@@ -143,29 +190,35 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
 
         # Work out the new value of unique_together, taking renames into
         # account
-        if override_uniques is None:
-            override_uniques = [
-                [rename_mapping.get(n, n) for n in unique]
-                for unique in model._meta.unique_together
-            ]
+        unique_together = [
+            [rename_mapping.get(n, n) for n in unique]
+            for unique in model._meta.unique_together
+        ]
 
         # Work out the new value for index_together, taking renames into
         # account
-        if override_indexes is None:
-            override_indexes = [
-                [rename_mapping.get(n, n) for n in index]
-                for index in model._meta.index_together
+        index_together = [
+            [rename_mapping.get(n, n) for n in index]
+            for index in model._meta.index_together
+        ]
+
+        indexes = model._meta.indexes
+        if delete_field:
+            indexes = [
+                index for index in indexes
+                if delete_field.name not in index.fields
             ]
 
         # Construct a new model for the new state
         meta_contents = {
             'app_label': model._meta.app_label,
             'db_table': model._meta.db_table,
-            'unique_together': override_uniques,
-            'index_together': override_indexes,
+            'unique_together': unique_together,
+            'index_together': index_together,
+            'indexes': indexes,
             'apps': apps,
         }
-        meta = type("Meta", tuple(), meta_contents)
+        meta = type("Meta", (), meta_contents)
         body['Meta'] = meta
         body['__module__'] = model.__module__
 
@@ -183,11 +236,11 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
 
         with altered_table_name(model, model._meta.db_table + "__old"):
             # Rename the old table to make way for the new
-            self.alter_db_table(model, temp_model._meta.db_table, model._meta.db_table)
-
-            # Create a new table with the updated schema. We remove things
-            # from the deferred SQL that match our table name, too
-            self.deferred_sql = [x for x in self.deferred_sql if temp_model._meta.db_table not in x]
+            self.alter_db_table(
+                model, temp_model._meta.db_table, model._meta.db_table,
+                disable_constraints=False,
+            )
+            # Create a new table with the updated schema.
             self.create_model(temp_model)
 
             # Copy data from the old table into the new table
@@ -212,27 +265,30 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
 
     def delete_model(self, model, handle_autom2m=True):
         if handle_autom2m:
-            super(DatabaseSchemaEditor, self).delete_model(model)
+            super().delete_model(model)
         else:
             # Delete the table (and only that)
             self.execute(self.sql_delete_table % {
                 "table": self.quote_name(model._meta.db_table),
             })
+            # Remove all deferred statements referencing the deleted table.
+            for sql in list(self.deferred_sql):
+                if isinstance(sql, Statement) and sql.references_table(model._meta.db_table):
+                    self.deferred_sql.remove(sql)
 
     def add_field(self, model, field):
         """
-        Creates a field on a model.
-        Usually involves adding a column, but may involve adding a
-        table instead (for M2M fields)
+        Create a field on a model. Usually involves adding a column, but may
+        involve adding a table instead (for M2M fields).
         """
         # Special-case implicit M2M tables
         if field.many_to_many and field.remote_field.through._meta.auto_created:
             return self.create_model(field.remote_field.through)
-        self._remake_table(model, create_fields=[field])
+        self._remake_table(model, create_field=field)
 
     def remove_field(self, model, field):
         """
-        Removes a field from a model. Usually involves deleting a column,
+        Remove a field from a model. Usually involves deleting a column,
         but for M2Ms may involve deleting a table.
         """
         # M2M fields are a special case
@@ -246,45 +302,31 @@ class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
             # It might not actually have a column behind it
             if field.db_parameters(connection=self.connection)['type'] is None:
                 return
-            self._remake_table(model, delete_fields=[field])
+            self._remake_table(model, delete_field=field)
 
     def _alter_field(self, model, old_field, new_field, old_type, new_type,
                      old_db_params, new_db_params, strict=False):
-        """Actually perform a "physical" (non-ManyToMany) field update."""
+        """Perform a "physical" (non-ManyToMany) field update."""
         # Alter by remaking table
-        self._remake_table(model, alter_fields=[(old_field, new_field)])
-
-    def alter_index_together(self, model, old_index_together, new_index_together):
-        """
-        Deals with a model changing its index_together.
-        Note: The input index_togethers must be doubly-nested, not the single-
-        nested ["foo", "bar"] format.
-        """
-        self._remake_table(model, override_indexes=new_index_together)
-
-    def alter_unique_together(self, model, old_unique_together, new_unique_together):
-        """
-        Deals with a model changing its unique_together.
-        Note: The input unique_togethers must be doubly-nested, not the single-
-        nested ["foo", "bar"] format.
-        """
-        self._remake_table(model, override_uniques=new_unique_together)
+        self._remake_table(model, alter_field=(old_field, new_field))
+        # Rebuild tables with FKs pointing to this field if the PK type changed.
+        if old_field.primary_key and new_field.primary_key and old_type != new_type:
+            for rel in new_field.model._meta.related_objects:
+                if not rel.many_to_many:
+                    self._remake_table(rel.related_model)
 
     def _alter_many_to_many(self, model, old_field, new_field, strict):
-        """
-        Alters M2Ms to repoint their to= endpoints.
-        """
+        """Alter M2Ms to repoint their to= endpoints."""
         if old_field.remote_field.through._meta.db_table == new_field.remote_field.through._meta.db_table:
             # The field name didn't change, but some options did; we have to propagate this altering.
             self._remake_table(
                 old_field.remote_field.through,
-                alter_fields=[(
+                alter_field=(
                     # We need the field that points to the target model, so we can tell alter_field to change it -
                     # this is m2m_reverse_field_name() (as opposed to m2m_field_name, which points to our model)
                     old_field.remote_field.through._meta.get_field(old_field.m2m_reverse_field_name()),
                     new_field.remote_field.through._meta.get_field(new_field.m2m_reverse_field_name()),
-                )],
-                override_uniques=(new_field.m2m_field_name(), new_field.m2m_reverse_field_name()),
+                ),
             )
             return
 
